@@ -1,5 +1,5 @@
 /*
- * Copyright 2009-2019 NVIDIA Corporation.  All rights reserved.
+ * Copyright 2009-2022 NVIDIA Corporation.  All rights reserved.
  *
  * NOTICE TO LICENSEE:
  *
@@ -67,10 +67,10 @@
 #define NVJPEG_MAX_COMPONENT 4
 
 // nvjpeg version information
-#define NVJPEG_VER_MAJOR 11
-#define NVJPEG_VER_MINOR 5
-#define NVJPEG_VER_PATCH 3
-#define NVJPEG_VER_BUILD 48
+#define NVJPEG_VER_MAJOR 12
+#define NVJPEG_VER_MINOR 3
+#define NVJPEG_VER_PATCH 5
+#define NVJPEG_VER_BUILD 92
 
 /* nvJPEG status enums, returned by nvJPEG API */
 typedef enum
@@ -85,9 +85,24 @@ typedef enum
     NVJPEG_STATUS_ARCH_MISMATCH                 = 7,
     NVJPEG_STATUS_INTERNAL_ERROR                = 8,
     NVJPEG_STATUS_IMPLEMENTATION_NOT_SUPPORTED  = 9,
+    NVJPEG_STATUS_INCOMPLETE_BITSTREAM          = 10
 } nvjpegStatus_t;
 
 
+// Enums for EXIF Orientation
+typedef enum nvjpegExifOrientation
+{
+    NVJPEG_ORIENTATION_UNKNOWN = 0,
+    NVJPEG_ORIENTATION_NORMAL = 1,
+    NVJPEG_ORIENTATION_FLIP_HORIZONTAL = 2,
+    NVJPEG_ORIENTATION_ROTATE_180 = 3,
+    NVJPEG_ORIENTATION_FLIP_VERTICAL = 4,
+    NVJPEG_ORIENTATION_TRANSPOSE = 5,
+    NVJPEG_ORIENTATION_ROTATE_90 = 6,
+    NVJPEG_ORIENTATION_TRANSVERSE = 7,
+    NVJPEG_ORIENTATION_ROTATE_270 = 8
+} nvjpegExifOrientation_t;
+
 // Enum identifies image chroma subsampling values stored inside JPEG input stream
 // In the case of NVJPEG_CSS_GRAY only 1 luminance channel is encoded in JPEG input stream
 // Otherwise both chroma planes are present
@@ -123,8 +138,10 @@ typedef enum
     NVJPEG_OUTPUT_RGBI        = 5, 
     // convert to interleaved BGR and write to 1-st channel of nvjpegImage_t
     NVJPEG_OUTPUT_BGRI        = 6,
+    // 16 bit unsigned output in interleaved format
+    NVJPEG_OUTPUT_UNCHANGEDI_U16 = 7,
     // maximum allowed value
-    NVJPEG_OUTPUT_FORMAT_MAX  = 6  
+    NVJPEG_OUTPUT_FORMAT_MAX  = 7
 } nvjpegOutputFormat_t;
 
 // Parameter of this type specifies what type of input user provides for encoding
@@ -140,14 +157,19 @@ typedef enum
 // NVJPEG_BACKEND_DEFAULT    : default value
 // NVJPEG_BACKEND_HYBRID     : uses CPU for Huffman decode
 // NVJPEG_BACKEND_GPU_HYBRID : uses GPU assisted Huffman decode. nvjpegDecodeBatched will use GPU decoding for baseline JPEG bitstreams with
-//                             interleaved scan when batch size is bigger than 100
+//                             interleaved scan when batch size is bigger than 50
 // NVJPEG_BACKEND_HARDWARE   : supports baseline JPEG bitstream with single scan. 410 and 411 sub-samplings are not supported
+// NVJPEG_BACKEND_GPU_HYBRID_DEVICE : nvjpegDecodeBatched will support bitstream input on device memory
+// NVJPEG_BACKEND_HARDWARE_DEVICE   : nvjpegDecodeBatched will support bitstream input on device memory
 typedef enum 
 {
     NVJPEG_BACKEND_DEFAULT = 0,
     NVJPEG_BACKEND_HYBRID  = 1,
     NVJPEG_BACKEND_GPU_HYBRID = 2,
-    NVJPEG_BACKEND_HARDWARE = 3
+    NVJPEG_BACKEND_HARDWARE = 3,
+    NVJPEG_BACKEND_GPU_HYBRID_DEVICE = 4,
+    NVJPEG_BACKEND_HARDWARE_DEVICE = 5,
+    NVJPEG_BACKEND_LOSSLESS_JPEG = 6,
 } nvjpegBackend_t;
 
 // Currently parseable JPEG encodings (SOF markers)
@@ -157,7 +179,8 @@ typedef enum
 
     NVJPEG_ENCODING_BASELINE_DCT                            = 0xc0,
     NVJPEG_ENCODING_EXTENDED_SEQUENTIAL_DCT_HUFFMAN         = 0xc1,
-    NVJPEG_ENCODING_PROGRESSIVE_DCT_HUFFMAN                 = 0xc2
+    NVJPEG_ENCODING_PROGRESSIVE_DCT_HUFFMAN                 = 0xc2,
+    NVJPEG_ENCODING_LOSSLESS_HUFFMAN                        = 0xc3
 
 } nvjpegJpegEncoding_t;
 
@@ -173,6 +196,10 @@ typedef enum
 #define NVJPEG_FLAGS_HW_DECODE_NO_PIPELINE 1
 #define NVJPEG_FLAGS_ENABLE_MEMORY_POOLS   1<<1
 #define NVJPEG_FLAGS_BITSTREAM_STRICT  1<<2
+#define NVJPEG_FLAGS_REDUCED_MEMORY_DECODE  1<<3
+#define NVJPEG_FLAGS_REDUCED_MEMORY_DECODE_ZERO_COPY  1<<4
+#define NVJPEG_FLAGS_UPSAMPLING_WITH_INTERPOLATION 1<<5
+
 
 // Output descriptor.
 // Data that is written to planes depends on output format
@@ -210,6 +237,30 @@ typedef struct
     tPinnedFree pinned_free;
 } nvjpegPinnedAllocator_t;
 
+
+typedef int (*tDevMallocV2)(void* ctx, void **ptr, size_t size, cudaStream_t stream);
+
+typedef int (*tDevFreeV2)(void* ctx, void *ptr, size_t size, cudaStream_t stream);
+
+
+typedef int (*tPinnedMallocV2)(void* ctx, void **ptr, size_t size, cudaStream_t stream);
+
+typedef int (*tPinnedFreeV2)(void* ctx, void *ptr, size_t size, cudaStream_t stream);
+
+typedef struct
+{
+    tDevMallocV2 dev_malloc;
+    tDevFreeV2 dev_free;
+    void *dev_ctx;
+} nvjpegDevAllocatorV2_t;
+
+typedef struct
+{
+    tPinnedMallocV2 pinned_malloc;
+    tPinnedFreeV2 pinned_free;
+    void *pinned_ctx;
+} nvjpegPinnedAllocatorV2_t;
+
 // Opaque library handle identifier.
 struct nvjpegHandle;
 typedef struct nvjpegHandle* nvjpegHandle_t;
@@ -246,6 +297,12 @@ nvjpegStatus_t NVJPEGAPI nvjpegCreateEx(nvjpegBackend_t backend,
         unsigned int flags,
         nvjpegHandle_t *handle);
 
+nvjpegStatus_t NVJPEGAPI nvjpegCreateExV2(nvjpegBackend_t backend,
+        nvjpegDevAllocatorV2_t *dev_allocator,
+        nvjpegPinnedAllocatorV2_t *pinned_allocator,
+        unsigned int flags,
+        nvjpegHandle_t *handle);
+
 // Release the handle and resources.
 // IN/OUT     handle: instance handle to release 
 nvjpegStatus_t NVJPEGAPI nvjpegDestroy(nvjpegHandle_t handle);
@@ -273,6 +330,10 @@ nvjpegStatus_t NVJPEGAPI nvjpegSetPinnedMemoryPadding(size_t padding, nvjpegHand
 nvjpegStatus_t NVJPEGAPI nvjpegGetPinnedMemoryPadding(size_t *padding, nvjpegHandle_t handle);
 
 
+nvjpegStatus_t NVJPEGAPI nvjpegGetHardwareDecoderInfo(nvjpegHandle_t handle,
+        unsigned int* num_engines,
+        unsigned int* num_cores_per_engine);
+
 
 // Initalization of decoding state
 // IN         handle        : Library handle
@@ -387,6 +448,19 @@ nvjpegStatus_t NVJPEGAPI nvjpegDecodeBatchedPreAllocate(
           nvjpegChromaSubsampling_t chroma_subsampling,
           nvjpegOutputFormat_t output_format);
 
+
+// Allocates the internal buffers as a pre-allocation step
+// IN    handle          : Library handle
+// IN    jpeg_handle     : Decoded jpeg image state handle
+// IN    data            : jpeg bitstream containing huffman and quantization tables
+// IN    length          : bitstream size in bytes
+
+nvjpegStatus_t NVJPEGAPI nvjpegDecodeBatchedParseJpegTables(
+          nvjpegHandle_t handle,
+          nvjpegJpegState_t jpeg_handle,
+          const unsigned char *data,
+          const size_t length);
+
 /**********************************************************
 *                        Compression                      *
 **********************************************************/
@@ -487,6 +561,12 @@ nvjpegStatus_t NVJPEGAPI nvjpegBufferPinnedCreate(nvjpegHandle_t handle,
     nvjpegPinnedAllocator_t* pinned_allocator,
     nvjpegBufferPinned_t* buffer);
 
+nvjpegStatus_t NVJPEGAPI nvjpegBufferPinnedCreateV2(nvjpegHandle_t handle,
+    nvjpegPinnedAllocatorV2_t* pinned_allocator,
+    nvjpegBufferPinned_t* buffer);
+
+nvjpegStatus_t NVJPEGAPI nvjpegBufferPinnedResize(nvjpegBufferPinned_t buffer, size_t size, cudaStream_t stream);
+
 nvjpegStatus_t NVJPEGAPI nvjpegBufferPinnedDestroy(nvjpegBufferPinned_t buffer);
 
 struct nvjpegBufferDevice;
@@ -496,6 +576,12 @@ nvjpegStatus_t NVJPEGAPI nvjpegBufferDeviceCreate(nvjpegHandle_t handle,
     nvjpegDevAllocator_t* device_allocator, 
     nvjpegBufferDevice_t* buffer);
 
+nvjpegStatus_t NVJPEGAPI nvjpegBufferDeviceCreateV2(nvjpegHandle_t handle,
+    nvjpegDevAllocatorV2_t* device_allocator,
+    nvjpegBufferDevice_t* buffer);
+
+nvjpegStatus_t NVJPEGAPI nvjpegBufferDeviceResize(nvjpegBufferDevice_t buffer, size_t size, cudaStream_t stream);
+
 nvjpegStatus_t NVJPEGAPI nvjpegBufferDeviceDestroy(nvjpegBufferDevice_t buffer);
 
 // retrieve buffer size and pointer - this allows reusing buffer when decode is not needed
@@ -540,7 +626,13 @@ nvjpegStatus_t NVJPEGAPI nvjpegJpegStreamParseHeader(
     nvjpegHandle_t handle,
     const unsigned char *data, 
     size_t length,
-    nvjpegJpegStream_t jpeg_stream);    
+    nvjpegJpegStream_t jpeg_stream);
+
+nvjpegStatus_t NVJPEGAPI nvjpegJpegStreamParseTables(
+    nvjpegHandle_t handle,
+    const unsigned char *data,
+    size_t length,
+    nvjpegJpegStream_t jpeg_stream);
 
 nvjpegStatus_t NVJPEGAPI nvjpegJpegStreamGetJpegEncoding(
     nvjpegJpegStream_t jpeg_stream,
@@ -561,8 +653,13 @@ nvjpegStatus_t NVJPEGAPI nvjpegJpegStreamGetComponentDimensions(
     unsigned int* width,
     unsigned int* height);
 
+nvjpegStatus_t NVJPEGAPI nvjpegJpegStreamGetExifOrientation(
+    nvjpegJpegStream_t jpeg_stream,
+    nvjpegExifOrientation_t *orientation_flag);
 
-
+nvjpegStatus_t NVJPEGAPI nvjpegJpegStreamGetSamplePrecision(
+    nvjpegJpegStream_t jpeg_stream,
+    unsigned int *precision);
 
 // if encoded is 1 color component then it assumes 4:0:0 (NVJPEG_CSS_GRAY, grayscale)
 // if encoded is 3 color components it tries to assign one of the known subsamplings
@@ -606,6 +703,11 @@ nvjpegStatus_t NVJPEGAPI nvjpegDecodeParamsSetScaleFactor(
     nvjpegDecodeParams_t decode_params,
     nvjpegScaleFactor_t scale_factor);
 
+// set the orientation flag to the decode parameters
+nvjpegStatus_t NVJPEGAPI nvjpegDecodeParamsSetExifOrientation(
+    nvjpegDecodeParams_t decode_params,
+    nvjpegExifOrientation_t orientation);
+
 ///////////////////////////////////////////////////////////////////////////////////
 // Decoder helper functions //
 ///////////////////////////////////////////////////////////////////////////////////
